/*****************************************************************************/
/*                                                                           */
/* File: FurTexture.cpp                                                      */
/*                                                                           */
/*****************************************************************************/

#include "FurTexture.h"

#include <d3d9.h>
#include <d3dx9.h>

const float INV_RAND_MAX = 1.0 / (RAND_MAX + 1);
inline float rnd(float max=1.0) { return max * INV_RAND_MAX * rand(); }
inline float rnd(float min, float max) { return min + (max - min) * INV_RAND_MAX * rand(); }

extern LPDIRECT3DDEVICE9       g_pd3dDevice;


/*****************************************************************************/
/*                                                                           */
/* FurTexture()                                                              */
/*                                                                           */
/*****************************************************************************/
FurTexture::FurTexture()
{
	m_Size		= 0;
	m_NumLayers	= 0;
}//End FurTexture()


/*****************************************************************************/
/*                                                                           */
/* ~FurTexture()                                                             */
/*                                                                           */
/*****************************************************************************/
FurTexture::~FurTexture()
{
	for (UINT layer = 0; layer < m_NumLayers; layer++)
		m_Textures[layer]->Release();
}//End ~FurTexture()


/*****************************************************************************/
/*                                                                           */
/* Create(..)                                                                */
/*                                                                           */
/*****************************************************************************/
bool FurTexture::Create(UINT seed, UINT size, UINT num)
{
	srand(seed);
	m_Size      = size;
	m_NumLayers = num;

	D3DXCOLOR *data = new D3DXCOLOR[m_NumLayers*m_Size*m_Size];
	#define DATA(layer, x, y) data[m_Size*m_Size*(layer) + m_Size*(y) + (x)]

	for (UINT x = 0; x < m_Size; x++)
	for (UINT y = 0; y < m_Size; y++)
	{
		UINT layer;
		D3DXCOLOR color;
		float t;

		// hair color
		color.r = rnd(1.0, 1.0);
		color.g = rnd(0.6, 0.8);
		color.b = rnd(0.2, 0.5);
		if (rnd() < 0.15) color *= rnd(0.2, 0.4);
		else              color *= rnd(0.7, 1.0);

		for (layer = 0; layer < m_NumLayers; layer++)
		{
			// lower layer is darker
			t = max(0, 1 - 2 * float(layer) / (m_NumLayers-1));
			DATA(layer, x, y) = color * float(1 - 0.7 * powf(t, 1.5));
		}

		// length of the hair
		UINT length = 1 + UINT(m_NumLayers * pow(rnd(), 3));
		length = min(m_NumLayers, length);

		for (layer = 0; layer < length; layer++)
		{
			// tip of the hair is semi-transparent
			t = max(0, -1 + 2 * float(layer+1) / length);
			DATA(layer, x, y).a = 1 - 0.85 * pow(t, 2);
		}
	}

	// Create all the textures
	for (UINT layer = 0; layer < m_NumLayers; layer++)
	{
		DWORD *pixels = new DWORD[m_Size*m_Size];
		DWORD *p = pixels;
		D3DXCOLOR *pColor = &DATA(layer, 0, 0);
		for(UINT i=0; i<m_Size*m_Size; i++)
			*p++ = DWORD(*pColor++);

		/*
			m_pd3dDevice->CreateTexture(512, 512, 1  // this won't work with LockRect(..)
                                 , D3DUSAGE_RENDERTARGET, D3DFMT_X8R8G8B8
                                 , D3DPOOL_DEFAULT, &m_pTexture );
		*/

		//IDirect3DTexture9* pTexture;

		HRESULT hr = g_pd3dDevice->CreateTexture(m_Size, m_Size, 1
                                 , 0, D3DFMT_A8R8G8B8
                                 , D3DPOOL_MANAGED, &m_Textures[layer], NULL );

		/*
		HRESULT hr = g_pd3dDevice->CreateTexture(m_Size, m_Size, 1
                                 , 0, D3DFMT_X8R8G8B8
                                 , D3DPOOL_MANAGED, &pTexture , NULL);
		*/

		/*
		DAMM!   A FEW HOURS OF MY LIFE WASTED!
		Well people, its always important to read the small print in docs,
		nearly 3am, and spent ages wondering why my LockRect(..) funciton wouldn't work,
		well its seems, you can't use 'D3DPOOL_DEFAULT' if your going to use LockRect(..)
		it will always fail.
		But then you'll also have problems if you try and use D3DUSAGE_RENDERTARGET mixed
		with other ones etc...worth looking into...bit of trial and error etc, but it
		works!
		*/

		IDirect3DTexture9* pTexture = m_Textures[layer];

		D3DLOCKED_RECT d3dlr;
		//HRESULT hr2 = pTexture->LockRect( 0, &d3dlr, 0, 0 );
		//HRESULT hr2 = pTexture->LockRect(0,&d3dlr,NULL,D3DLOCK_READONLY );
		HRESULT hr2 = pTexture->LockRect(0,&d3dlr,NULL,0 );

		// pitch value should be h * w * 4 = 128 x 4 = 

		//m_Textures[layer]->LockRect( 0, &d3dlr, 0, 0 );
		DWORD * pDst = (DWORD *)d3dlr.pBits;
		int DPitch = d3dlr.Pitch/4;
		DWORD * pSrc = (DWORD *)pixels;
		int SPitch = d3dlr.Pitch/4;	// bad assumption, but lets assume there the same pitch!
		
		for (int i=0; i<m_Size/*height*/; ++i)
			for (int j=0; j<m_Size/*width*/; ++j)
				pDst[i*DPitch + j] = pSrc[i*SPitch + j];				
		
		pTexture->UnlockRect (0);

		delete [] pixels;
	}

	delete [] data;

	return true;
}//End Create(..)


/*****************************************************************************/
/*                                                                           */
/* GetTexture(..)                                                            */
/*                                                                           */
/*****************************************************************************/
void FurTexture::GetTexture(float layer, IDirect3DTexture9** pTexture) const
{
	int temp = (int)floor(m_NumLayers * max(0.0, min(0.9999, layer)));

	*pTexture = m_Textures[temp];
}//End GetTexture(..)

